#!/usr/local/bin/python2.7

import __builtin__
import os
import string
import sys
import time

import pygame

def _(*message):
	return string.join(message)

__builtin__.__dict__["_"] = _

from bombz import appdir, bitmaps, board, controls, gamefont, joy, joymenu
from bombz import key, keymenu, levels, menu, options, optsmenu, sounds
from bombz import particles, pusher, userdir

FPS = 50
TIME_MARGIN = 0
INTRO_FONT_SIZE = 20
INTRO_COLOUR = (80, 255, 80)
GET_READY_COLOUR = (255, 255, 40)
INTRO_MARGIN = 8


class OutcomeMenu(menu.Menu):
	""" Menu to be shown when game is finished. """
	def __init__(self, caption_strings, items,
			logo = None, title = None, caption_color = None,
			fgcolor = None, bgcolor = None, tlcolor = None):
		menu.Menu.__init__(self, title, logo, bgcolor, tlcolor)
		i = 0
		if not caption_color:
			caption_color = (220, 255, 240)
		for c in caption_strings:
			self.append(menu.CaptionMenuItem(c, caption_color))
			i += 1
		self.first_highlightable = i
		for mi in items:
			self.append(mi)


class OutcomeMenuManager:
	def __init__(self, logo = None, caption_color = None,
			fgcolor = None, bgcolor = None, tlcolor = None):
		self.logo = logo
		self.caption_color = caption_color
		self.fgcolor = fgcolor
		self.bgcolor = bgcolor
		self.tlcolor = tlcolor
		self.menus = [None] * pusher.N_OUTCOMES
	
	def create_menu(self, outcome):
		if outcome == pusher.COMPLETED:
			self.menus[outcome] = OutcomeMenu( \
					[_("Congratulations! You completed"),
					_("the level in time.")],
					[self.__next_level_item(), self.__replay_level_item(), 
						self.__main_menu_item(), self.__quit_item()],
					self.logo)
		elif outcome == pusher.COMPLETED_ALL:
			self.menus[outcome] = OutcomeMenu( \
					["Congratulations! You completed all the levels",
						"at this difficulty."],
					[self.__replay_level_item(), 
						self.__main_menu_item(), self.__quit_item()],
					self.logo)
		elif outcome == pusher.COMPLETED_OUT_OF_TIME_EASY:
			self.menus[outcome] = OutcomeMenu( \
					[("You completed the level but you'll have to"),
					_("do it faster than that to beat the Medium"),
					_("and hard levels.")],
					[self.__next_level_item(), self.__replay_level_item(), 
						self.__main_menu_item(), self.__quit_item()],
					self.logo)
		elif outcome == pusher.COMPLETED_OUT_OF_TIME_MEDIUM:
			self.menus[outcome] = OutcomeMenu( \
					[("You completed the level but you took"),
					_("too long. Try to do it faster now.")],
					[self.__replay_level_item(), 
						self.__main_menu_item(), self.__quit_item()],
					self.logo)
		elif outcome == pusher.BLOWN_UP:
			self.menus[outcome] = OutcomeMenu( \
					[("Bang! You got caught in an explosion.")],
					[self.__replay_level_item(), 
						self.__main_menu_item(), self.__quit_item()],
					self.logo)
		elif outcome == pusher.OUT_OF_TIME:
			self.menus[outcome] = OutcomeMenu( \
					[("Time's up!")],
					[self.__replay_level_item(), 
						self.__main_menu_item(), self.__quit_item()],
					self.logo)

	def run_menu(self, screen, outcome):
		if not self.menus[outcome]:
			self.create_menu(outcome)
		menu = self.menus[outcome]
		menu.highlighted = menu.first_highlightable
		menu.show(screen)
		self.menu = menu
		menu.event_loop()
		return self.result

	# Result codes
	NEXT_LEVEL = 0
	REPLAY_LEVEL = 1
	MAIN_MENU = 2
	QUIT = 3

	def __next_level_item(self):
		return menu.MenuItem(self.__choose_next_level,
				_("Play next level"), self.fgcolor, self.bgcolor)
	
	def __choose_next_level(self):
		self.result = self.__class__.NEXT_LEVEL
		self.menu.exit()

	def __replay_level_item(self):
		return menu.MenuItem(self.__choose_replay_level,
				_("Play this level again"), self.fgcolor, self.bgcolor)
	
	def __choose_replay_level(self):
		self.result = self.__class__.REPLAY_LEVEL
		self.menu.exit()

	def __main_menu_item(self):
		return menu.MenuItem(self.__choose_main_menu,
				_("Main menu"), self.fgcolor, self.bgcolor)
	
	def __choose_main_menu(self):
		self.result = self.__class__.MAIN_MENU
		self.menu.exit()

	def __quit_item(self):
		return menu.MenuItem(self.__choose_quit,
				_("Quit"), self.fgcolor, self.bgcolor)
	
	def __choose_quit(self):
		self.result = self.__class__.QUIT
		self.menu.exit()



class Bombz:
	def __init__(self):
		pygame.mixer.pre_init(48000)
		pygame.init()
		self.options = options.Options()
		particles.init(self.options)
		self.o_fullscreen = options.Option("fullscreen", self.options, 0)
		self.open_display()
		sounds.init()
		self.time_bg = None
		pygame.display.set_caption(_("Bombz"))
		key.init(self.options)
		joy.init_joysticks(self.options)
		self.load_difficulty()
		self.levels = levels.LevelSet()
		self.level = self.levels.get_next_level(self.difficulty)
		self.pusher = pusher.Pusher()
		self.main_menu = self.create_main_menu()
		self.pause_menu = self.create_pause_menu()
		self.difficulty_menu = self.create_difficulty_menu()
		self.outcome_menu = \
				OutcomeMenuManager(bitmaps.get_bitmap("biglogo.png"))
		self.show_main_menu()
		self.main_menu.event_loop()
	
	def __get_choose_level_data(self):
		""" Returns [ [level_indexes], [level_names], current] """
		names = []
		if self.difficulty:
			levs = self.levels.get_completed_levels(self.difficulty)
		else:
			levs = []
		next = self.levels.get_next_level(self.difficulty)
		if next == None:
			next = 0
		levs.append(next)
		for l in levs:
			names.append(self.levels.get_level_name(l))
		index = None
		if self.level != None:
			for i in range(len(levs)):
				if levs[i] == self.level:
					index = i
		if index == None:
			index = len(levs) - 1
		return [levs, names, index]
	
	def show_main_menu(self):
		i = self.resume_mitem
		if os.path.exists(self.difficulty_filename()):
			i.sensitive = True
		else:
			i.sensitive = False
			self.main_menu.highlighted = 0
		i = self.choose_level_mitem
		choosable = self.__get_choose_level_data()
		i.value_values = choosable[0]
		i.value_names = choosable[1]
		i.index = choosable[2]
		self.main_menu.show(self.screen)
	
	def open_display(self):
		if int(self.o_fullscreen.get()):
			opts = pygame.FULLSCREEN | pygame.DOUBLEBUF | pygame.HWSURFACE
		else:
			opts = 0
		self.screen = pygame.display.set_mode((640, 480), opts)
		pygame.mouse.set_visible(opts == 0)
		gamefont.init()
		self.digit_width, self.digit_height = gamefont.get_digit_size()
		self.digit_origin = [self.digit_width + TIME_MARGIN,
				self.screen.get_height() - self.digit_height - 4 - TIME_MARGIN]
		self.intro_font = pygame.font.Font(appdir.get_font_filename(),
				INTRO_FONT_SIZE)
	
	def create_main_menu(self):
		logo = bitmaps.get_bitmap("biglogo.png")
		self.opts_menu = optsmenu.OptionsMenu(self.options, logo,
				self.o_fullscreen, self.fullscreen_toggled)
		m = menu.Menu(_("Main Menu"), logo)
		m.append(menu.MenuItem(self.new_game, _("New game")))
		i = menu.MenuItem(self.resume, _("Resume game"))
		m.append(i)
		self.resume_mitem = i
		choosable = self.__get_choose_level_data()
		self.choose_level_mitem = \
				menu.ActivatableOptionMenuItem(self.choose_level,
						_("Choose level"), choosable[2],
						choosable[1], choosable[0], align_around = 240)
		m.append(self.choose_level_mitem)
		m.append(menu.MenuItem(self.open_opts_menu, _("Options")))
		m.append(menu.ExitMenuItem(_("Quit")))
		m.highlighted = 1
		return m

	def create_pause_menu(self):
		logo = bitmaps.get_bitmap("biglogo.png")
		m = menu.Menu(_("Pause Menu"), logo)
		m.append(menu.ExitMenuItem(_("Resume current game")))
		m.append(menu.MenuItem(self.pause_set_restart, _("Restart level")))
		m.append(menu.MenuItem(self.pause_quit_game, _("Quit to main menu")))
		m.append(menu.MenuItem(self.pause_quit_bombz, _("Quit Bombz")))
		return m
	
	def pause_quit_game(self):
		self.aborted = True
		self.pause_menu.exit()
	
	def pause_set_restart(self):
		self.aborted = True
		self.restart_level = True
		self.pause_menu.exit()
	
	def pause_quit_bombz(self):
		self.levels.failed(self.level)
		sys.exit(0)
	
	def create_difficulty_menu(self):
		logo = bitmaps.get_bitmap("biglogo.png")
		m = menu.Menu(_("Choose difficulty"), logo)
		m.append(menu.MenuItem(self.set_easy, _("Easy")))
		m.append(menu.MenuItem(self.set_medium, _("Medium")))
		m.append(menu.MenuItem(self.set_difficult, _("Difficult")))
		return m
	
	def fullscreen_toggled(self, value):
		self.o_fullscreen.set(value)
		self.open_display()
		self.opts_menu.show(self.screen)
	
	def render_screen(self, surface, blit = True, hide_detonated = False):
		self.board.RenderFloor(surface)
		self.pusher.RenderBlank(surface)
		self.board.RenderContent(surface, hide_detonated)
		self.pusher.RenderBody(surface)
		particles.process_all(surface)
		self.render_time(surface)
		if blit:
			self.screen.blit(surface, (0, 0))
			pygame.display.flip()
	
	def poll_kbd(self, any_key = False):
		" Returns 0 for no event, 1 for esc, 2 for other key or mouse click. "
		for event in pygame.event.get():
			if event.type == pygame.QUIT:
				sys.exit(0)
			elif event.type == pygame.MOUSEBUTTONDOWN:
				return 2
			else:
				ie = controls.interpret_event(event)
				if ie[3]:
					return 1
				if ie[0] and any_key:
					return 2
		return 0

	def show_caption(self, surface, text, y, colour):
		caption = self.intro_font.render(text, True, colour)
		caption_width = caption.get_width() + 2 * INTRO_MARGIN
		caption_height = caption.get_height() + 2 * INTRO_MARGIN
		caption_bg = pygame.Surface((caption_width,
			caption_height)).convert()
		caption_bg.set_alpha(160)
		x0 = int((640 - caption_width) / 2)
		surface.blit(caption_bg, (x0, y))
		surface.blit(caption, (x0 + INTRO_MARGIN, y + INTRO_MARGIN))

	def level_intro(self, surface, clock):
		flash_off = False
		for n in range(5 * FPS / 4):
			self.render_screen(surface, False, n % 2 == 0)
			level_name = levels.process_level_name(self.board.title)
			self.show_caption(surface, level_name, 180, INTRO_COLOUR)
			self.show_caption(surface, _("Get ready"), 300,
					GET_READY_COLOUR)
			self.screen.blit(surface, (0, 0))
			pygame.display.flip()
			poll = self.poll_kbd(True)
			if poll == 1:
				self.pause()
				if self.aborted:
					return False
			elif poll == 2:
				return True
			clock.tick(FPS / 4)
			self.board.FlashBombsInIntro(flash_off)
			flash_off = not flash_off
		return True

	def pause(self):
		paused_at = pygame.time.get_ticks()
		self.pause_menu.highlighted = 0
		self.pause_menu.show(self.screen)
		self.pause_menu.event_loop()
		self.start_time += pygame.time.get_ticks() - paused_at
	
	def play_loop(self):
		""" Return results are constants from stary of pusher module """
		surface = pygame.Surface(self.screen.get_size())
		clock = pygame.time.Clock()
		# Set start_time here in case pause during intro
		self.start_time = pygame.time.get_ticks()
		if not self.level_intro(surface, clock):
			return pusher.ABORTED
		if self.board.HasFused():
			sounds.fuse.play()
		self.start_time = pygame.time.get_ticks()
		n = 0
		flash_off = False
		while True:
			if self.poll_kbd() == 1:
				self.pause()
				if self.aborted:
					return pusher.ABORTED
			result = self.pusher.Update()
			if result:
				return result
			n += 1
			if n == 4:
				flash_off = not flash_off
				n = 0
			self.board.UpdateBombs(flash_off)
			self.time_left = self.time_limit - \
					(pygame.time.get_ticks() - self.start_time) / 1000
			if self.difficulty == levels.HARD and self.time_left <= 0:
				return pusher.OUT_OF_TIME
			self.render_screen(surface)
			clock.tick(FPS)
	
	def __blit_glyph(self, surf, glyph, x, width = None):
		if width:
			x += int((width - glyph.get_width()) / 2)
		y = self.digit_origin[1] + \
				int((self.digit_height - glyph.get_height()) / 2)
		surf.blit(glyph, (x, y))

	def __blit_digit(self, surf, digits, digit, x):
		glyph = digits[digit % 10]
		self.__blit_glyph(surf, glyph, x, self.digit_width)
	
	def render_time(self, surf):
		if self.time_left <= -1:
			return
		if self.time_left < 0:
			self.time_left = 0
		if self.time_left <= 7:
			last_pip = 8 - int(self.time_left)
			if last_pip > self.last_pip:
				if last_pip % 2:
					sounds.pip1.play()
				else:
					sounds.pip2.play()
				self.last_pip = last_pip
		mins = int(self.time_left / 60)
		secs = self.time_left % 60
		tsecs = int(secs / 10)
		secs %= 10
		x = self.digit_origin[0]
		if self.time_left < 10 and self.time_left > 0:
			red = (pygame.time.get_ticks() / 125) % 2 == 0
		elif self.time_left <= 15:
			red = True
		else:
			red = False
		if red:
			digits = gamefont.danger_digits
			colon = gamefont.danger_colon
		else:
			digits = gamefont.digits
			colon = gamefont.colon
		time_bg_width = 3 * self.digit_width + colon.get_width() + \
				TIME_MARGIN * 2
		if mins > 10:
			time_bg_width += self.digit_width
		time_bg_height = self.digit_height + TIME_MARGIN * 2
		if not self.time_bg or \
				self.time_bg.get_width() != time_bg_width or \
				self.time_bg.get_height() != time_bg_height:
			self.time_bg = pygame.Surface((time_bg_width,
				time_bg_height)).convert()
			self.time_bg.set_alpha(100)
		surf.blit(self.time_bg, (self.digit_origin[0] - TIME_MARGIN,
				self.digit_origin[1] - TIME_MARGIN))
		if mins > 10:
			self.__blit_digit(surf, digits, int(mins / 10), x)
			mins %= 10
			x += self.digit_width
		self.__blit_digit(surf, digits, mins, x)
		x += self.digit_width
		self.__blit_glyph(surf, colon, x)
		x += colon.get_width()
		self.__blit_digit(surf, digits, tsecs, x)
		x += self.digit_width
		self.__blit_digit(surf, digits, secs, x)

	def play(self):
		self.restart_level = True
		while self.restart_level:
			self.last_pip = 0
			self.restart_level = False
			self.aborted = False
			self.main_menu.showing = False
			self.board = board.Board( \
					self.levels.get_level_filename(self.level),
					self.difficulty == levels.EASY)
			self.pusher.NewBoard(self.board, self.levels.get_stats(self.level))
			self.time_limit = self.board.time_limit
			self.time_left = self.time_limit
			self.levels.start_attempt(self.level, self.difficulty,
					self.board.CountDetonators(), self.time_limit)
			result = self.play_loop()
			if result == pusher.COMPLETED and self.time_left <= 0:
				if self.difficulty == levels.EASY:
					result = pusher.COMPLETED_OUT_OF_TIME_EASY
				elif self.difficulty == levels.MEDIUM:
					result = pusher.COMPLETED_OUT_OF_TIME_MEDIUM
			if result <= pusher.COMPLETED_OUT_OF_TIME_EASY:
				self.levels.succeeded(self.level,
					self.board.CountDetonators(), self.time_left)
				self.main_menu.highlighted = 1
				self.levels.get_next_level(self.difficulty) 
				if self.levels.get_next_level(self.difficulty) == None:
					os.remove(self.difficulty_filename())
					result = pusher.COMPLETED_ALL
			else:
				self.levels.failed(self.level)
			particles.delete_all()
			if self.restart_level:
				continue
			if result == pusher.ABORTED:
				break
			result = self.outcome_menu.run_menu(self.screen, result)
			if result == OutcomeMenuManager.MAIN_MENU:
				self.restart_level = False
			else:
				self.restart_level = True
			if result == OutcomeMenuManager.NEXT_LEVEL:
				self.level = self.levels.get_next_level(self.difficulty)
			elif result == OutcomeMenuManager.QUIT:
				sys.exit(0)
		self.show_main_menu()
	
	def choose_level(self, level_index):
		self.level = level_index
		self.play()

	def resume(self):
		self.level = self.levels.get_next_level(self.difficulty)
		self.play()

	def new_game(self):
		self.difficulty_menu.highlighted = 1
		self.difficulty_menu.show(self.screen)
		self.difficulty_menu.event_loop()
		self.levels.reset_completed(self.difficulty)
		self.level = 0
		self.main_menu.highlighted = 0
		self.play()
	
	def set_easy(self):
		self.set_difficulty(levels.EASY)
		self.difficulty_menu.exit()

	def set_medium(self):
		self.set_difficulty(levels.MEDIUM)
		self.difficulty_menu.exit()

	def set_difficult(self):
		self.set_difficulty(levels.DIFFICULT)
		self.difficulty_menu.exit()

	def set_difficulty(self, difficulty):
		self.difficulty = difficulty
		fp = open(self.difficulty_filename(), 'w')
		print >>fp, difficulty
		fp.close()
	
	def load_difficulty(self):
		try:
			fp = open(self.difficulty_filename(), 'r')
			self.difficulty = int(fp.readline())
			fp.close()
		except:
			self.difficulty = levels.MEDIUM

	@staticmethod
	def difficulty_filename():
		return os.path.join(userdir.dirs.data(), "Difficulty")

	def open_opts_menu(self):
		self.main_menu.showing = False
		self.opts_menu.show(self.screen)
		self.opts_menu.event_loop()
		self.show_main_menu()


appdir.init(sys.argv[0])
userdir.dirs = userdir.UserDirs("realh.co.uk", "Bombz")
levels.load_default_set()
bombz = Bombz()
pygame.quit()
