import pygame
import random

from bombz import appdir, bitmaps, particles, sounds

TILE_SIZE = 32
COLUMNS = 20
ROWS = 15

FUSE_TICKS = 60
EXPLO_TICKS = 12
EASY_EXTRA_FUSE = 20
# EASY_EXTRA_FUSE must be multiple of 4 for flashing to work correctly

[ CELL_EMPTY,
CELL_EARTH0,
CELL_EARTH1,
CELL_EARTH2,
CELL_PICKET,
CELL_BOMB1,
CELL_BOMB2,
CELL_MATCH,
CELL_CHROME00,
CELL_CHROME01,
CELL_CHROME02,
CELL_CHROME03,
CELL_CHROME04,
CELL_CHROME05,
CELL_CHROME06,
CELL_CHROME07,
CELL_CHROME08,
CELL_CHROME09,
CELL_CHROME10,
CELL_CHROME11,
CELL_CHROME12,
CELL_CHROME13,
CELL_CHROME14,
CELL_CHROME15 ] = range(24)

CELL_BOMB1_FUSED_FIRST = CELL_CHROME15 + 1
CELL_BOMB1_FUSED_LAST = CELL_BOMB1_FUSED_FIRST + \
		FUSE_TICKS + EASY_EXTRA_FUSE - 1
CELL_BOMB2_FUSED_FIRST = CELL_BOMB1_FUSED_LAST + 1
CELL_BOMB2_FUSED_LAST = CELL_BOMB2_FUSED_FIRST + \
		FUSE_TICKS + EASY_EXTRA_FUSE - 1

CELL_EXPLO_FIRST = CELL_BOMB2_FUSED_LAST + 1
CELL_EXPLO_LAST = CELL_EXPLO_FIRST + EXPLO_TICKS - 1

# Cache explosion graphics
explo_bitmaps = []

def CacheExploBitmaps():
	global explo_bitmaps
	for n in range(EXPLO_TICKS):
		explo_bitmaps.append(bitmaps.get_bitmap("explo%d" % (n)))

def IsChrome(a):
	return a >= CELL_CHROME00 and a <= CELL_CHROME15

def IsUnfusedBomb(a):
	return a == CELL_BOMB1 or a == CELL_BOMB2

def IsPushable(a):
	return a == CELL_BOMB1 or a == CELL_BOMB2

def IsDestructable(a):
	return (a >= CELL_EARTH0 and a < CELL_EARTH2) \
			or a == CELL_PICKET or a == CELL_MATCH

def IsExplosion(a):
	return a >= CELL_EXPLO_FIRST and a <= CELL_EXPLO_LAST

def GetFloorBitmap(floor_bitmap = None):
	if not floor_bitmap:
		floor_bitmap = [bitmaps.get_bitmap("floor")]
	return floor_bitmap[0]

def FlashBitmap(bitmap, index, flash_off = False):
	if flash_off:
		return GetFloorBitmap()
	else:
		return bitmap

def GetBitmapForContent(content, flash_off = False):
	if content == CELL_EMPTY:
		return bitmaps.get_bitmap("floor")
	elif content >= CELL_EARTH0 and content <= CELL_EARTH2:
		return bitmaps.get_bitmap("earth%d" % (content - CELL_EARTH0))
	elif content == CELL_PICKET:
		return bitmaps.get_bitmap("picket")
	elif content == CELL_BOMB1:
		return bitmaps.get_bitmap("bomb1")
	elif content == CELL_BOMB2:
		return bitmaps.get_bitmap("bomb2")
	elif content == CELL_MATCH:
		return bitmaps.get_bitmap("match")
	elif content >= CELL_BOMB1_FUSED_FIRST \
			and content <= CELL_BOMB1_FUSED_LAST:
		return FlashBitmap(bitmaps.get_bitmap("bomb1"), \
				content - CELL_BOMB1_FUSED_FIRST, flash_off)
	elif content >= CELL_BOMB2_FUSED_FIRST \
			and content <= CELL_BOMB2_FUSED_LAST:
		return FlashBitmap(bitmaps.get_bitmap("bomb2"), \
				content - CELL_BOMB2_FUSED_FIRST, flash_off)
	elif content >= CELL_EXPLO_FIRST and content <= CELL_EXPLO_LAST:
		return explo_bitmaps[content - CELL_EXPLO_FIRST]
	else:
		return bitmaps.get_bitmap("chrome%02d" % (content - CELL_CHROME00))

def ClipX(x):
	if x < 0:
		x = 0
	elif x >= COLUMNS:
		x = COLUMNS - 1;
	return x

def ClipY(y):
	if y < 0:
		y = 0
	elif y >= ROWS:
		y = ROWS - 1;
	return y

def ClipRect(x0, y0, x1, y1):
	return [ClipX(x0), ClipY(y0), ClipX(x1), ClipY(y1)]

class Cell:
	def SetContent(self, content, flash_off = False):
		self.content = content
		self.bitmap = GetBitmapForContent(content, flash_off)

	def __init__(self, content = CELL_EMPTY):
		self.SetContent(content)

	def GetContent(self):
		return self.content


# A Cell which has knowledge of its position within a board
class CellInBoard(Cell):
	def __init__(self, board, x, y, content = CELL_EMPTY):
		Cell.__init__(self, content)
		self.board = board
		self.x = x
		self.y = y

	def GetNeighbour(self, dx, dy):
		x = self.x + dx
		y = self.y + dy
		if x >= 0 and x < COLUMNS and y >= 0 and y < ROWS:
			return self.board.GetCellAt(x, y)
		else:
			return None

	def GetNeighbourContent(self, dx, dy):
		cell = self.GetNeighbour(dx, dy)
		if cell:
			return cell.GetContent()
		else:
			return None
	
	def RenderBitmap(self, surface, bitmap):
		surface.blit(bitmap, (self.x * TILE_SIZE, self.y * TILE_SIZE))
	
	def RenderFloor(self, surface):
		if self.content >= CELL_EARTH0 and self.content <= CELL_EARTH2:
			self.RenderBitmap(surface, self.bitmap)
		else:
			self.RenderBitmap(surface, GetFloorBitmap())
	
	# Returns TRUE if content is CELL_EXPLO_FIRST
	def RenderContent(self, surface, hide_detonated = False):
		if self.content == CELL_EXPLO_FIRST:
			return True
		if hide_detonated and self.content >= CELL_BOMB1_FUSED_FIRST:
			return False
		if self.content > CELL_EARTH2:
			self.RenderBitmap(surface, self.bitmap)
		return False


class Board:
	def Reset(self):
		self.title = _("Unnamed")
		self.cells = []
		self.nbombs = 0
		for y in range(ROWS):
			row = []
			for x in range(COLUMNS):
				row.append(CellInBoard(self, x, y))
			self.cells.append(row)

	def GetCellAt(self, x, y):
		if x < 0 or x >= COLUMNS or y < 0 or y >= ROWS:
			return None
		return self.cells[y][x]

	# Unbombable walls in original were in solid blocks, but with new
	# graphics they look much better as a railing around an empty space,
	# so empty out those spaces. Relies on all chrome initially being CHROME15
	def HollowOutChrome(self):
		cells = []
		for y in range(ROWS):
			row = []
			for x in range(COLUMNS):
				cell = self.GetCellAt(x, y)
				content = cell.GetContent()
				if content == CELL_CHROME15:
					if cell.GetNeighbourContent(-1, 0) == CELL_CHROME15 \
							and cell.GetNeighbourContent(1, 0) == \
								CELL_CHROME15 \
							and cell.GetNeighbourContent(0, -1) == \
								CELL_CHROME15 \
							and cell.GetNeighbourContent(0, 1) == CELL_CHROME15:
						content = CELL_EMPTY
				row.append(CellInBoard(self, x, y, content))
			cells.append(row)
		self.cells = cells
	
	# Once chrome areas have been hollowed out we can convert the outlines
	# to all their different shapes. See mappings.txt
	def ShapeChrome(self):
		cells = []
		for y in range(ROWS):
			row = []
			for x in range(COLUMNS):
				cell = self.GetCellAt(x, y)
				content = cell.GetContent()
				left = cell.GetNeighbourContent(-1, 0) == CELL_CHROME15
				right = cell.GetNeighbourContent(1, 0) == CELL_CHROME15
				above = cell.GetNeighbourContent(0, -1) == CELL_CHROME15
				below = cell.GetNeighbourContent(0, 1) == CELL_CHROME15
				if content == CELL_CHROME15:
					if left:
						if right:
							if above:
								if below:
									# Chrome on all sides
									content = CELL_CHROME08
								else:
									# Free below only
									content = CELL_CHROME11
							else:
								if below:
									# Free above only
									content = CELL_CHROME13
								else:
									# Free above and below
									content = CELL_CHROME01
						else:
							if above:
								if below:
									# Free to right only
									content = CELL_CHROME12
								else:
									# Free to right and below
									content = CELL_CHROME05
							else:
								if below:
									# Free above and to right
									content = CELL_CHROME02
								else:
									# Chrome on left only
									content = CELL_CHROME09
					else:
						if right:
							if above:
								if below:
									# Free to left only
									content = CELL_CHROME14
								else:
									# Free below and to left
									content = CELL_CHROME04
							else:
								if below:
									# Chrome below and to right
									content = CELL_CHROME00
								else:
									# Chrome to right only
									content = CELL_CHROME07
						else:
							if above:
								if below:
									# Chrome above and below
									content = CELL_CHROME03
								else:
									# Chrome above only
									content = CELL_CHROME10
							else:
								if below:
									# Chrome below only
									content = CELL_CHROME06
								else:
									# Free on all sides
									content = CELL_CHROME15
				row.append(CellInBoard(self, x, y, content))
			cells.append(row)
		self.cells = cells
	
	# If we have two parallel rows or columns they will be unnecessarily
	# linked with T sections ie we can replace:
	#
	#    ===========         ===========
	#    ==+=+=+=+==         ===========
	#      | | | |     with
	#    ==+=+=+=+==         ===========
	#    ===========         ===========
	#
	# for a nicer look
	def DisconnectTs(self):
		for y in range(ROWS):
			for x in range(COLUMNS):
				cell1 = self.GetCellAt(x, y)
				if cell1.GetContent() == CELL_CHROME13:
					cell2 = cell1.GetNeighbour(0, 1)
					if cell2 and cell2.GetContent() == CELL_CHROME11:
						cell1.SetContent(CELL_CHROME01)
						cell2.SetContent(CELL_CHROME01)
				elif cell1.GetContent() == CELL_CHROME14:
					cell2 = cell1.GetNeighbour(1, 0)
					if cell2 and cell2.GetContent() == CELL_CHROME12:
						cell1.SetContent(CELL_CHROME03)
						cell2.SetContent(CELL_CHROME03)

	def RandomiseEarthAndBombs(self):
		for y in range(ROWS):
			for x in range(COLUMNS):
				cell = self.GetCellAt(x, y)
				if cell.GetContent() == CELL_EARTH0:
					cell.SetContent(random.randint(CELL_EARTH0, CELL_EARTH2))
				if cell.GetContent() == CELL_BOMB1:
					cell.SetContent(random.randint(CELL_BOMB1, CELL_BOMB2))
				elif cell.GetContent() == self.bomb1_fused_first:
					if random.randint(0, 2):
						content = self.bomb1_fused_first
					else:
						content = self.bomb2_fused_first
					cell.SetContent(content)

	def Beautify(self):
		# This should ensure earth looks consistent each time a level is loaded
		random.seed(self.title)
		self.HollowOutChrome()
		self.ShapeChrome()
		self.DisconnectTs()
		self.RandomiseEarthAndBombs()

	def LoadLevel(self, filename, easy = False):
		self.fused_x = -1
		self.fused_y = -1
		self.set_easy(easy)
		fd = open(filename, 'r')
		for y in range(ROWS):
			line = fd.readline()
			#print "Line %02d: %s" % (y, line)
			for x in range(COLUMNS):
				cell = self.GetCellAt(x, y)
				c = line[x]
				if c == ' ':
					cell.SetContent(CELL_EMPTY)
				elif c == '.':
					cell.SetContent(CELL_EARTH0)
				elif c == 'B':
					cell.SetContent(CELL_PICKET)
				elif c == 'O':
					cell.SetContent(CELL_BOMB1)
					self.nbombs += 1
				elif c == '0':
					self.fused_x = x
					self.fused_y = y
					cell.SetContent(self.bomb1_fused_first)
					self.nbombs += 1
				elif c == 'D':
					cell.SetContent(CELL_MATCH)
				elif c == 'S':
					cell.SetContent(CELL_EMPTY)
					self.start_x = x
					self.start_y = y
				elif c == 'X':
					cell.SetContent(CELL_CHROME15)
		self.time_limit = int(fd.readline())
		password = fd.readline()	# Password is obsolete feature
		self.title = fd.readline()
		self.Beautify()

	def __init__(self, level = None, easy = False): 
		global explo_bitmaps
		if not explo_bitmaps:
			CacheExploBitmaps()
		self.start_x = 0
		self.start_y = 0
		self.Reset()
		if level:
			self.LoadLevel(level, easy)
		else:
			self.set_easy(easy)
	
	def set_easy(self, easy):
		self.chain_fuse1 = CELL_BOMB1_FUSED_LAST
		self.chain_fuse2 = CELL_BOMB2_FUSED_LAST
		self.bomb1_fused_first = CELL_BOMB1_FUSED_FIRST
		self.bomb2_fused_first = CELL_BOMB2_FUSED_FIRST
		if easy:
			self.chain_fuse1 -= 1
			self.chain_fuse2 -= 1
		else:
			self.bomb1_fused_first += EASY_EXTRA_FUSE
			self.bomb2_fused_first += EASY_EXTRA_FUSE
	
	def RenderFloor(self, surface):
		for y in range(ROWS):
			for x in range(COLUMNS):
				self.GetCellAt(x, y).RenderFloor(surface)

	def RenderContent(self, surface, hide_detonated = False):
		explo = False
		for y in range(ROWS):
			for x in range(COLUMNS):
				if self.GetCellAt(x, y).RenderContent(surface, hide_detonated):
					explo = True
		if explo:
			for y in range(ROWS):
				for x in range(COLUMNS):
					if self.GetCellAt(x, y).GetContent() == CELL_EXPLO_FIRST:
						surface.blit(explo_bitmaps[0], \
								((x - 1) * TILE_SIZE, (y - 1) * TILE_SIZE))

	def GetStartPos(self):
		return (self.start_x, self.start_y)

	def GetContentAt(self, x, y):
		cell = self.GetCellAt(x, y)
		if cell:
			return cell.GetContent()
		else:
			return None

	def SetContentAt(self, x, y, content):
		self.GetCellAt(x, y).SetContent(content)
	
	def AddParticles(self, x, y):
		particles.create_flurry(x * TILE_SIZE + TILE_SIZE / 2,
				y * TILE_SIZE + TILE_SIZE / 2)

	# Adds a new explosion at cell x, y adjacent to an old explosion at x0, y0.
	# Needs to know x0, y0, because if x, y is after it in the UpdateBombs
	# loop it needs to be FUSED_LAST - 2 instead of FUSED_LAST - 1
	def ChainExplode(self, x0, y0, x, y, cell, content):
		if y > y0 or (x > x0 and y == y0):
			content -= 1
		cell.SetContent(content)
	
	def ExplodeAt(self, x0, y0, cell):
		sounds.bomb.play()
		cell.SetContent(CELL_EXPLO_FIRST)
		self.AddParticles(x0, y0)
		[x1, y1, x2, y2] = ClipRect(x0 - 1, y0 - 1, x0 + 1, y0 + 1)
		for y in range(y1, y2 + 1):
			for x in range(x1, x2 + 1):
				if x == x0 and y == y0:
					continue
				cell = self.GetCellAt(x, y)
				content = cell.GetContent()
				if IsDestructable(content):
					cell.SetContent(CELL_EMPTY)
				elif content == CELL_BOMB1 \
						or (content >= CELL_BOMB1_FUSED_FIRST \
						and content < self.chain_fuse1):
					self.ChainExplode(x0, y0, x, y, cell, self.chain_fuse1)
				elif content == CELL_BOMB2 \
						or (content >= CELL_BOMB2_FUSED_FIRST \
						and content < self.chain_fuse2):
					self.ChainExplode(x0, y0, x, y, cell, self.chain_fuse2)

	def UpdateBombs(self, flash_off = False):
		for y in range(ROWS):
			for x in range(COLUMNS):
				cell = self.GetCellAt(x, y)
				content = cell.GetContent()
				if (content >= CELL_BOMB1_FUSED_FIRST \
						and content < CELL_BOMB1_FUSED_LAST) \
						or (content >= CELL_BOMB2_FUSED_FIRST \
						and content < CELL_BOMB2_FUSED_LAST):
					cell.SetContent(content + 1, flash_off)
				elif content == CELL_BOMB1_FUSED_LAST \
						or content == CELL_BOMB2_FUSED_LAST:
					self.ExplodeAt(x, y, cell)
					self.nbombs -= 1
				elif content >= CELL_EXPLO_FIRST and content < CELL_EXPLO_LAST:
					cell.SetContent(content + 1)
					# Check whether pusher caught in explosion
				elif content == CELL_EXPLO_LAST:
					cell.SetContent(CELL_EMPTY)
	
	def FlashBombsInIntro(self, flash_off):
		for y in range(ROWS):
			for x in range(COLUMNS):
				cell = self.GetCellAt(x, y)
				content = cell.GetContent()
				if content == self.bomb1_fused_first or \
						content == self.bomb2_fused_first:
					# This refreshes bitmap
					cell.SetContent(content, flash_off)

	def GotAllBombs(self):
		return self.nbombs == 0

	def CountDetonators(self):
		n = 0
		for row in self.cells:
			for c in row:
				if c.GetContent() == CELL_MATCH:
					n += 1
		return n

	def HasFused(self):
		n = 0
		for row in self.cells:
			for c in row:
				c = c.GetContent()
				if (c >= CELL_BOMB1_FUSED_FIRST and \
							c <= CELL_BOMB1_FUSED_LAST) or \
						(c >= CELL_BOMB2_FUSED_FIRST and \
							c <= CELL_BOMB2_FUSED_LAST):
					return True
		return False

