import pygame
import sys

from bombz import options

try:
	from curses.ascii import ismeta, iscntrl
except:
	def ismeta(c):
		return ord(c) > 128

	def iscntrl(c):
		return ord(c) < 32 or ord(c) == 127

joysticks = []
active_joystick = None
active_index = -1
o_active_joystick = None
joy_names = {}

MAX_NAME_LEN = 35

def clamp(value, min, max):
	if value < min:
		value = min
	elif value > max:
		value = max
	return value

class Joystick:
	""" Joystick wrapped up in extra configuration stuff etc. """
	def __init__(self, id, opts):
		""" id is a pygame joystick id.
		opts is an Options object. """
		self.id = id
		self.joystick = pygame.joystick.Joystick(id)
		global joy_names
		self.name = None
		n = self.get_name()
		if joy_names.has_key(n):
			self.instance = joy_names[n] + 1
		else:
			self.instance = 0
		joy_names[n] = self.instance
		self.options = opts
		self.config_name = n.replace(' ', '_').replace('=', '_') \
				+ '.' + str(self.instance)
		global joysticks
		joysticks.append(self)
		self.horiz_axis_val = 0
		self.vert_axis_val = 0
		self.setup_options()
		self.button_pressed = None

	def setup_options(self):
		opts = self.options
		self.o_hat = options.Option(self.opt_name('hat'), opts, -1)
		self.hat = int(self.o_hat.get())
		self.o_horiz_axis = options.Option(self.opt_name('horiz_axis'),
				opts, 0)
		self.horiz_axis = int(self.o_horiz_axis.get())
		self.o_vert_axis = options.Option(self.opt_name('vert_axis'),
				opts, 1)
		self.vert_axis = int(self.o_vert_axis.get())
		self.o_ok_button = options.Option(self.opt_name('ok_button'),
				opts, 0)
		self.ok_button = int(self.o_ok_button.get())
		self.o_back_button = options.Option(self.opt_name('back_button'),
				opts, 1)
		self.back_button = int(self.o_back_button.get())
		self.o_threshold = options.Option(self.opt_name('axis_threshold'),
				opts, 0.4)
		self.threshold = float(self.o_threshold.get())
	
	def reduce_spaces(self, name):
		while name.find('  ') != -1:
			name = name.replace('  ', ' ')
		return name.strip()

	def remove_word(self, name, word, rpt_lwr = True):
		if len(name) > MAX_NAME_LEN:
			name = name.replace(word, '')
			name = self.reduce_spaces(name)
			if rpt_lwr:
				name = self.remove_word(name, word.lower(), False)
		return name
	
	def get_name(self):
		if not self.name:
			name = self.joystick.get_name()
			# Names can be a bit long, so shorten
			# Anything top-bit-set is probably something like (R) so replace
			for n in range(len(name)):
				if ismeta(name[n]) or iscntrl(name[n]):
					name = name[:n] + ' ' + name[n + 1:]
			name = self.reduce_spaces(name)
			# My MS Sidewinder USB pad has "Microsoft" in it twice
#			words = name.split()
#			if words[0] == words[1]:
#				name = self.reduce_spaces(name.replace(words[0], '', 1))
#			self.name = name
			# If name is still too long, try getting rid of extra guff
			# Most of this is targetted at the Sidewinder
			name = name.replace('Plug & Play', 'PnP')
			if len(name) > MAX_NAME_LEN:
				name = name.replace("Microsoft", "MS")
			name = self.remove_word(name, "Controller")
			name = self.remove_word(name, "Game ")
			if len(name) > MAX_NAME_LEN:
				# Get rid of first word, probably brand,
				# still left with unique model name
				s = name.find(' ')
				if s != -1:
					name = name[s + 1:]
			self.name = name
		return self.name

	def opt_name(self, subname):
		""" Converts an option subname eg 'enabled' into full option name. """
		return 'joy.' + self.config_name + '.' + subname

	def get_direction(self):
		""" Gets the direction indicated by this joystick
		as bools [l, r, u, d] """
		result = [False, False, False, False]
		if self.hat != -1:
			hat = self.joystick.get_hat(self.hat)
			if hat[0] < 0:
				result[0] = True
			elif hat[0] > 0:
				result[1] = True
			if hat[1] > 0:
				result[2] = True
			elif hat[1] < 0:
				result[3] = True
		if self.horiz_axis != -1:
			x = self.joystick.get_axis(self.horiz_axis)
			if x < -self.threshold:
				result[0] = True
			elif x > self.threshold:
				result[1] = True
		if self.vert_axis != -1:
			y = self.joystick.get_axis(self.vert_axis)
			if y < -self.threshold:
				result[2] = True
			elif y > self.threshold:
				result[3] = True
		return result

def init_joysticks(opts):
	global active_joystick
	global active_index
	global o_active_joystick
	o_active_joystick = options.Option("joystick", opts, -1)
	n_sticks = pygame.joystick.get_count()
	for i in range(n_sticks):
		j = Joystick(i, opts)
		if j.get_name() == o_active_joystick.get():
			active_index = i
			active_joystick = j
			j.joystick.init()


def interpret_event(e):
	""" Returns [is_joy_event (bool), [x, y], action, back] """
	if not active_joystick:
		return [False, [0, 0], False, False]
	if e.type == pygame.JOYAXISMOTION and e.joy == active_joystick.id:
		result = [0, 0]
		if e.axis == active_joystick.vert_axis:
			if e.value < -active_joystick.threshold \
					and active_joystick.vert_axis_val \
							>= -active_joystick.threshold:
				result[1] = -1
			elif e.value > active_joystick.threshold \
					and active_joystick.vert_axis_val \
							<= active_joystick.threshold:
				result[1] = 1
			active_joystick.vert_axis_val = e.value
			return [True, result, False, False]
		elif e.axis == active_joystick.horiz_axis:
			if e.value < -active_joystick.threshold \
					and active_joystick.horiz_axis_val \
							>= -active_joystick.threshold:
				result[0] = -1
			elif e.value > active_joystick.threshold \
					and active_joystick.horiz_axis_val \
							<= active_joystick.threshold:
				result[0] = 1
			active_joystick.horiz_axis_val = e.value
			return [True, result, False, False]
	elif e.type == pygame.JOYHATMOTION and e.joy == active_joystick.id:
		return [True, [e.value[0], -e.value[1]], False, False]
	elif e.type == pygame.JOYBUTTONDOWN and e.joy == active_joystick.id:
		if e.button == active_joystick.ok_button:
			return [True, [0, 0], True, False]
		elif e.button == active_joystick.back_button:
			return [True, [0, 0], False, True]
	return [False, [0, 0], False, False]

def get_direction():
	if active_joystick:
		return active_joystick.get_direction()
	return [False, False, False, False]
