import pygame
import sys

from bombz import joy
from bombz import menu

ALIGN_AROUND = 200

class JoystickPicker(menu.OptionMenuItem):
	def __init__(self, jmopts):
		self.jmopts = jmopts
		joynames = [_("None")]
		for j in joy.joysticks:
			n = j.get_name()
			i = joy.joy_names[n]
			if i > 0:
				n = "%s (%d)" % (n, i)
			joynames.append(n)
		menu.OptionMenuItem.__init__(self, jmopts.joy_picked_handler,
				_("Controller"), joy.active_index + 1,
				joynames, range(pygame.joystick.get_count() + 1), ALIGN_AROUND)


class PropertyPicker(menu.OptionMenuItem):
	" Base class to be overridden to choose axes, hats and buttons from menu. "
	def __init__(self, jmopts, label, default_value,
			value_format, range_method, prop_name):
		self.jmopts = jmopts
		self.value_format = value_format
		self.range_method = range_method
		self.prop_name = prop_name
		self.option_name = 'o_' + prop_name
		self.activated = False
		[names, values] = self.__get_names_and_values()
		if joy.active_joystick:
			value = int(getattr(joy.active_joystick, self.option_name).get())
		else:
			value = default_value
		menu.OptionMenuItem.__init__(self, self.handler, label,
				value + 1, names, values, ALIGN_AROUND)
		self.__set_sensitive()
	
	def __get_names_and_values(self):
		names = [_("Disabled")]
		values = [-1]
		if joy.active_joystick:
			for n in range(getattr(joy.active_joystick.joystick,
					self.range_method)()):
				names.append(self.value_format % n)
				values.append(n)
		return [names, values]

	def __set_sensitive(self):
		if joy.active_joystick:
			self.sensitive = int(getattr(joy.active_joystick.joystick, 
					self.range_method)()) > 0
		else:
			self.sensitive = False

	def joystick_changed(self):
		[names, indexes] = self.__get_names_and_values()
		if joy.active_joystick:
			value = int(getattr(joy.active_joystick, self.option_name).get())
		else:
			value = 0
		self.value_names = names
		self.value_values = indexes
		self.index = value + 1
		self.__set_sensitive()
		self.updated()

	def handler(self, value):
		getattr(joy.active_joystick, self.option_name).set(value)
		setattr(joy.active_joystick, self.prop_name, value)
	
	def get_value_to_display(self, index):
		if self.activated:
			return " ??? "
		else:
			return menu.OptionMenuItem.get_value_to_display(self, index)

	def activate(self):
		self.activated = True
		self.menu.redraw()
		while True:
			event = pygame.event.wait()
			if event.type == pygame.QUIT:
				sys.exit(0)
			elif event.type == pygame.KEYDOWN \
				and event.key == pygame.K_ESCAPE:
					break
			else:
				value = self.value_from_event(event)
				if value is not None:
					for i in range(len(self.value_values)):
						if self.value_values[i] == value:
							self.index = i
					self.handler(value)
					break
		self.activated = False
		self.menu.redraw()

	def value_from_event(self, event):
		""" Override. """
		return None



class ButtonPicker(PropertyPicker):
	def __init__(self, jmopts, label, prop_name):
		self.button_down = None
		PropertyPicker.__init__(self, jmopts, label, -1,
				_("Button %d"), "get_numbuttons", prop_name)

	def value_from_event(self, event):
		if event.type == pygame.JOYBUTTONDOWN \
				and joy.active_joystick.id == event.joy:
			self.button_down = event.button
		elif event.type == pygame.JOYBUTTONUP \
				and self.button_down == event.button \
				and joy.active_joystick.id == event.joy:
			return event.button
		else:
			return None

	def handler(self, value, check_clashes = True):
		# If this button is already assigned to another action, swap
		if check_clashes and value != None:
			old_val = int(getattr(joy.active_joystick, self.prop_name))
			self.jmopts.swap_clash(self, old_val, value)
		PropertyPicker.handler(self, value)



class AxisPicker(PropertyPicker):
	def __init__(self, jmopts, label, default_value, prop_name):
		PropertyPicker.__init__(self, jmopts, label, default_value,
				_("Axis %d"), "get_numaxes", prop_name)

	def value_from_event(self, event):
		if event.type == pygame.JOYAXISMOTION \
				and abs(event.value) > joy.active_joystick.threshold \
				and joy.active_joystick.id == event.joy:
			return event.axis
		else:
			return None


class HatPicker(PropertyPicker):
	def __init__(self, jmopts):
		self.button_down = None
		PropertyPicker.__init__(self, jmopts, _("Hat"), 0,
				_("Hat %d"), "get_numhats", "hat")

	def value_from_event(self, event):
		if event.type == pygame.JOYHATMOTION \
				and joy.active_joystick.id == event.joy:
			return event.hat
		else:
			return None

MIN_THRESHOLD = 0.1
MAX_THRESHOLD = 0.9
THRESHOLD_RANGE = MAX_THRESHOLD - MIN_THRESHOLD

class Sensitivity(menu.SliderMenuItem):
	def __init__(self, jmopts):
		start_value = self.joy_to_slider()
		menu.SliderMenuItem.__init__(self, _("Sensitivity"),
				0, THRESHOLD_RANGE * 10, 1, start_value,
				self.sensitivity_updated, ALIGN_AROUND)
	
	def joy_to_slider(self, value = None):
		if value is None:
			if joy.active_joystick:
				value = joy.active_joystick.threshold
			else:
				value = 0.4
		return int((MAX_THRESHOLD + MIN_THRESHOLD - value) * 10)
	
	def slider_to_joy(self, value = None):
		if value is None:
			value = self.value
		return MAX_THRESHOLD - float(value) / 10

	def sensitivity_updated(self, value):
		value = self.slider_to_joy(value)
		joy.active_joystick.o_threshold.set(value)
		joy.active_joystick.threshold = value
	
	def __set_sensitive(self):
		if joy.active_joystick:
			self.sensitive = True
		else:
			self.sensitive = False

	def joystick_changed(self):
		if joy.active_joystick:
			self.value = self.joy_to_slider()
			self.sensitive = True
		else:
			self.value = self.joy_to_slider(0.6)
			self.sensitive = False
		self.updated()


class JoystickMenuOptions:
	def __init__(self, opts):
		self.opts = opts
		self.m_pick_joy = JoystickPicker(self)
		self.m_horiz_axis = AxisPicker(self,
				_("Horizontal axis"), 0, "horiz_axis")
		self.m_vert_axis = AxisPicker(self,
				_("Vertical axis"), 1, "vert_axis")
		self.m_sensitivity = Sensitivity(self)
		self.m_hat = HatPicker(self)
		self.m_ok_button = ButtonPicker(self, _("OK button"), "ok_button")
		self.m_back_button = ButtonPicker(self, _("Back button"), "back_button")

	def joy_picked_handler(self, new_index):
		""" Remember new_index is +1 """
		if joy.active_joystick:
			joy.active_joystick.joystick.quit()
		joy.active_index = new_index - 1
		if new_index:
			joy.active_joystick = joy.joysticks[new_index - 1]
			joy.active_joystick.joystick.init()
			joy.o_active_joystick.set(joy.active_joystick.get_name())
		else:
			joy.active_joystick = None
			joy.o_active_joystick.set("None")
		# Avoid updating screen for each item; defer to end
		showing = self.menu.showing
		self.menu.showing = False
		for m in (self.m_horiz_axis, 
				self.m_vert_axis, 
				self.m_sensitivity, 
				self.m_hat, 
				self.m_ok_button, 
				self.m_back_button):
			m.joystick_changed()
		self.menu.showing = showing

	def create_menu(self, logo = None):
		m = menu.Menu(_("Configure game controller"), logo)
		self.menu = m
		m.append(self.m_pick_joy)
		m.append(self.m_horiz_axis)
		m.append(self.m_vert_axis)
		m.append(self.m_sensitivity)
		m.append(self.m_hat)
		m.append(self.m_ok_button)
		m.append(self.m_back_button)
		m.append(menu.ExitMenuItem())
		return m

	def swap_clash(self, button, old_val, new_val):
		for m in (self.m_ok_button, self.m_back_button):
			if m != button and int(getattr(joy.active_joystick, m.prop_name)) \
					== new_val:
				m.index = old_val + 1
				m.handler(old_val, False)
				m.updated()
				break


