import os
import sys

from bombz import appdir, userdir

try:
	from curses.ascii import isalpha, isdigit
except:
	def isalpha(c):
		return (ord(c) >= ord('A') and ord(c) <= ord('Z')) or \
				(ord(c) >= ord('a') and ord(c) <= ord('z'))
	def isdigit(c):
		return ord(c) >= ord('0') and ord(c) <= ord('9')

NOT_DONE = 0
EASY = 1
MEDIUM = 2
HARD = 4

def progress_dir():
	dir = os.path.join(userdir.dirs.data(), "Progress")
	userdir.ensure_dir(dir)
	return dir

def stats_dir():
	dir = os.path.join(userdir.dirs.data(), "Stats")
	userdir.ensure_dir(dir)
	return dir

def process_level_name(s):
	""" Returns new version of s stripped and with all non-initial letters
	converted to lower case. """
	s = s.strip()
	s2 = ""
	for n in range(len(s)):
		if n > 0 and (isalpha(s[n - 1]) or s[n - 1] == "'"):
			s2 += s[n].lower()
		else:
			s2 += s[n]
	return s2

def new_average(old_avg, new_n, new_value):
	return old_avg * (float(new_n - 1) / float(new_n)) + \
			float(new_value) / float(new_n)

N_STATS_ELEMENTS = 8

class Statistics:
	""" Gathers statistics about levels to help determine how difficult they
	are. """
	def __init__(self, array = None):
		if not array:
			array = [0] * N_STATS_ELEMENTS
		self.from_list(array)

	def from_list(self, a):
		self.failed_attempts = int(a[0])
		self.successful_attempts = int(a[1])
		self.failed_difficulties = int(a[2])
		self.successful_difficulties = int(a[3])
		self.detonators = int(a[4])
		self.avg_detonators_left = float(a[5])
		self.avg_moves = float(a[6])
		self.avg_ratio_time_left = float(a[7])

	def to_list(self):
		return [self.failed_attempts, self.successful_attempts,
				self.failed_difficulties, self.successful_difficulties,
				self.detonators, self.avg_detonators_left,
				self.avg_moves, self.avg_ratio_time_left]
	
	def start_attempt(self, difficulty, detonators, time_limit):
		self.difficulty = difficulty
		self.detonators = detonators
		self.moves = 0	# Actually changes of direction
		self.time_limit = time_limit
	
	def move_made(self, n = 1):
		" Actually changes of direction "
		self.moves += n
	
	def failed(self):
		self.failed_attempts += 1
		self.failed_difficulties |= self.difficulty
	
	def succeeded(self, time_left, detonators_left):
		self.successful_attempts += 1
		self.successful_difficulties |= self.difficulty
		self.avg_detonators_left = new_average(self.avg_detonators_left,
				self.successful_attempts, detonators_left)
		self.avg_moves = new_average(self.avg_moves,
				self.successful_attempts, self.moves)
		self.avg_ratio_time_left = new_average(self.avg_ratio_time_left,
				self.successful_attempts,
				float(time_left) / float(self.time_limit))



class LevelSet:
	def __init__(self, dir = None):
		if not dir:
			dir = appdir.get_levels_dir()
		self.dir = dir
		self.levels = []
		self.level_names = {}
		self.init_cache()
		self.load_progress()
		self.load_stats()
	
	def init_cache(self):
		self.levels = []
		self.level_names = {}
		dir = self.dir
		cache_file = os.path.join(userdir.dirs.cache(), os.path.basename(dir))
		# Check whether the cache needs (re)building
		if os.path.exists(cache_file):
			cache_mtime = os.stat(cache_file).st_mtime
			if os.stat(dir).st_mtime - cache_mtime > 0:
				build_cache = True
			else:
				build_cache = False
				for f in os.listdir(dir):
					f = os.path.join(dir, f)
					if os.stat(f).st_mtime - cache_mtime > 0:
						build_cache = True
						break
		else:
			build_cache = True
		if build_cache:
			files = os.listdir(dir)
			files.sort()
			for f in files:
				pathname = os.path.join(dir, f)
				if isdigit(f[0]) and os.path.isfile(pathname):
					fp = open(pathname, 'r')
					lines = fp.readlines()
					fp.close()
					level_name = process_level_name(lines[-1])
					lev = [f, level_name, NOT_DONE, None]
					self.levels.append(lev)
					self.level_names[level_name] = lev
			fp = open(cache_file, 'w')
			for l in self.levels:
				fp.write("%s %s\n" % (l[0], l[1]))
			fp.close()
		else:
			fp = open(cache_file, 'r')
			lines = fp.readlines()
			fp.close()
			for l in lines:
				f, level_name = l.split(' ', 1)
				level_name = level_name.strip()
				lev = [f, level_name, NOT_DONE, None]
				self.levels.append(lev)
				self.level_names[level_name] = lev
	
	def progress_filename(self):
		return os.path.join(progress_dir(), os.path.basename(self.dir))
	
	def load_progress(self):
		filename = self.progress_filename()
		if not os.path.exists(filename):
			return
		fp = open(filename, 'r')
		lines = fp.readlines()
		fp.close()
		for l in lines:
			completed, level_name = l.split(' ', 1)
			level_name = process_level_name(level_name)
			lev = self.level_names.get(level_name)
			if not lev:
				print >> sys.stderr, \
						_('Level named "%s" not known') % level_name
				continue
			lev[2] = int(completed)
	
	def save_progress(self):
		fp = open(self.progress_filename(), 'w')
		for l in self.levels:
			if l[2] > 0:
				fp.write("%d %s\n" % (l[2], l[1]))
		fp.close()
	
	def stats_filename(self):
		return os.path.join(stats_dir(), os.path.basename(self.dir))
	
	def load_stats(self):
		filename = self.stats_filename()
		if not os.path.exists(filename):
			return
		fp = open(filename, 'r')
		lines = fp.readlines()
		fp.close()
		for l in lines:
			stats = l.split(' ', N_STATS_ELEMENTS)
			level_name = process_level_name(stats[N_STATS_ELEMENTS])
			lev = self.level_names.get(level_name)
			if not lev:
				print >> sys.stderr, \
						_('Level named "%s" not known') % level_name
				continue
			lev[3] = Statistics(stats[:-1])
	
	def save_stats(self):
		fp = open(self.stats_filename(), 'w')
		for l in self.levels:
			if l[3]:
				for a in l[3].to_list():
					fp.write("%s " % str(a))
				fp.write(l[1] + '\n')
		fp.close()
	
	def get_next_level(self, difficulty):
		""" Returns index of first level not completed at this difficulty
		or higher. Result is None if all completed. """
		for n in range(len(self.levels)):
			if self.levels[n][2] < difficulty:
				return n
		else:
			return None
	
	def get_level_filename(self, i):
		return os.path.join(self.dir, self.levels[i][0])
	
	def get_level_name(self, i):
		return self.levels[i][1]
	
	def reset_completed(self, difficulty):
		""" Marks all levels as not completed at this difficulty or higher.
		If they've been completed at a lower difficulty that will be retained.
		"""
		for l in self.levels:
			l[2] &= difficulty - 1
		self.save_progress()
	
	def get_stats(self, i):
		lev = self.levels[i]
		if not lev[3]:
			lev[3] = Statistics()
		return lev[3]

	def start_attempt(self, i, difficulty, detonators, time_limit):
		self.difficulty = difficulty
		stats = self.get_stats(i)
		stats.start_attempt(difficulty, detonators, time_limit)

	def failed(self, i):
		stats = self.get_stats(i)
		stats.failed()
		self.save_stats()

	def succeeded(self, i, time_left, detonators_left):
		stats = self.get_stats(i)
		self.levels[i][2] |= self.difficulty
		stats.succeeded(time_left, detonators_left)
		self.save_progress()
		self.save_stats()
	
	def get_completed_levels(self, difficulty):
		""" Returns list of indexes. """
		completed = []
		for n in range(len(self.levels)):
			if self.levels[n][2] >= difficulty:
				completed.append(n)
		return completed


	
default_set = None

def load_default_set():
	default_set = LevelSet()
