import os
import pygame
import sys

from bombz import appdir
from bombz import bitmaps
from bombz import controls
from bombz import customevents

TOP_MARGIN = 32
LOGO_MARGIN = 12
LOGO_SPACING = 20
LINE_SPACING = 8
FONT_SIZE = 18
TITLE_FONT_SIZE = 24
TITLE_SPACING = 16
CENTRE_GAP = 20

_fonts = {}

def _render_text(text, fgcolor, bgcolor, font_size):
	font = _fonts.get(font_size)
	if not font:
		font = pygame.font.Font(appdir.get_font_filename(), font_size)
		_fonts[font_size] = font
	return font.render(text, True, fgcolor, bgcolor)

def render_text_at(dsurf, text, x, y, fgcolor, bgcolor, font_size):
	fsurf = _render_text(text, fgcolor, bgcolor, font_size)
	dsurf.blit(fsurf, (x, y))
	return [fsurf.get_width(), fsurf.get_height()]

def render_text_centered_at(dsurf, text, x, y, width,
		fgcolor, bgcolor, font_size, fill = False):
	fsurf = _render_text(text, fgcolor, bgcolor, font_size)
	if fill:
		pygame.draw.rect(dsurf, bgcolor, (x, y, width, fsurf.get_height()))
	x += (width - fsurf.get_width()) / 2
	dsurf.blit(fsurf, (x, y))
	return [fsurf.get_width(), fsurf.get_height()]

def render_centered_text(dsurf, text, y, fgcolor, bgcolor, font_size):
	return render_text_centered_at(dsurf, text, 0, y, dsurf.get_width(),
			fgcolor, bgcolor, font_size)

class Menu:
	""" logo is a pygame surface """
	def __init__(self, title = None, logo = None,
			bgcolor = None, tlcolor = None):
		if not bgcolor:
			bgcolor = (0, 0, 0)
		if not tlcolor:
			tlcolor = (70, 255, 100)

		self.title = title
		self.logo = logo
		self.bgcolor = bgcolor
		self.tlcolor = tlcolor

		if self.logo:
			self.top = LOGO_MARGIN + self.logo.get_height() + LOGO_SPACING
		else:
			self.top = TOP_MARGIN
		self.highlighted = 0
		self.items = []
		self.showing = False
	
	def append(self, item):
		self.items.append(item)
		item.menu = self
	
	def redraw(self):
		if not self.items[self.highlighted].sensitive:
			h = self.highlighted - 1
			while not self.items[h].sensitive and h > -1:
				h -= 1
			if h != -1:
				self.highlighted = h
			else:
				h = self.highlighted + 1
				while not self.items[h].sensitive \
						and h < len(self.items):
					h += 1
				if h != len(self.items):
					self.highlighted = h
		surf = self.surface.copy()
		surf.fill(self.bgcolor)
		if self.logo:
			x = (surf.get_width() - self.logo.get_width()) / 2
			surf.blit(self.logo, (x, LOGO_MARGIN))
		y = self.top
		if self.title:
			render_centered_text(surf, self.title, y,
					self.tlcolor, self.bgcolor, TITLE_FONT_SIZE)
			y += TITLE_FONT_SIZE + TITLE_SPACING
		for i in range(len(self.items)):
			h = self.items[i].render(surf, y, i == self.highlighted)
			y += h
			h -= FONT_SIZE
			dy = LINE_SPACING - h
			if dy > 0:
				y += dy
		self.surface.blit(surf, (0, 0))
		pygame.display.flip()
	
	def show(self, surface):
		self.surface = surface
		self.redraw()
		self.showing = True
	
	def event_loop(self):
		while True:
			event = pygame.event.wait()
			hli = self.items[self.highlighted]
			ctrl_event = controls.interpret_event(event)
			if event.type == pygame.QUIT:
				sys.exit(0)
			elif ctrl_event[0]:
				if ctrl_event[3]:
					return
				elif ctrl_event[1][1] == -1 and self.highlighted > 0:
					h = self.highlighted - 1
					while not self.items[h].sensitive and h > -1:
						h -= 1
					if h != -1:
						self.highlighted = h
					self.redraw()
				elif ctrl_event[1][1] == 1 \
						and self.highlighted < len(self.items) -1:
					h = self.highlighted + 1
					while not self.items[h].sensitive \
							and h < len(self.items):
						h += 1
					if h != len(self.items):
						self.highlighted = h
					self.redraw()
				elif ctrl_event[2]:
					hli.activate()
				elif ctrl_event[1][0] == 1:
					hli.inc()
				elif ctrl_event[1][0] == -1:
					hli.dec()
			elif event.type == pygame.USEREVENT \
					and event.code == customevents.EXIT_MENU:
				self.showing = False
				return
			else:
				hli.event_handler(event)

	def exit(self):
		pygame.event.post(pygame.event.Event(pygame.USEREVENT,
			{ 'code': customevents.EXIT_MENU} ))


class MenuItem:
	""" Basic menu item, single piece of text. Handler is the function to call
	    when activated. """
	def __init__(self, handler, label, fgcolor = None, bgcolor = None,
			hlfg = None, hlbg = None):
		self.menu = None
		if not fgcolor:
			fgcolor = (240, 240, 100)
		if not bgcolor:
			bgcolor = (0, 0, 0)
		if not hlfg:
			hlfg = (0, 0, 0)
		if not hlbg:
			hlbg = (180, 120, 40)
		self.fgcolor = fgcolor
		self.bgcolor = bgcolor
		self.hlfg = hlfg
		self.hlbg = hlbg
		self.label = label
		self.handler = handler
		self.sensitive = True

	def render(self, surf, y, highlighted):
		# Returns height of rendered line
		if highlighted:
			fgcolor = self.hlfg
			bgcolor = self.hlbg
		else:
			if self.sensitive:
				fgcolor = self.fgcolor
			else:
				fgcolor = self.hlbg
			bgcolor = self.bgcolor
		return render_centered_text(surf, self.label, y,
				fgcolor, bgcolor, FONT_SIZE)[1]
	
	def updated(self):
		if self.menu and self.menu.showing:
			self.menu.redraw()

	def activate(self):
		self.handler()
	def inc(self):
		pass
	def dec(self):
		pass
	
	def event_handler(self, event):
		" If the parent menu doesn't recognise the event it will call this. "
		pass


class CaptionMenuItem(MenuItem):
	""" Can never be highlighted """
	def __init__(self, label, fgcolor = None, bgcolor = None):
		MenuItem.__init__(self, None, label, fgcolor, bgcolor)
		self.hlbg = self.fgcolor
		self.sensitive = False


class ExitMenuItem(MenuItem):
	""" Special MenuItem to exit from the menu. """
	def __init__(self, label = None, fgcolor = None, bgcolor = None,
			hlfg = None, hlbg = None):
		if not label:
			label = _("Back")
		MenuItem.__init__(self, self.handler, label,
				fgcolor, bgcolor, hlfg, hlbg)
	def handler(self):
		self.menu.exit()


class LabelledMenuItem(MenuItem):
	""" A menu item with a separate label and value. """
	def __init__(self, label, align_around = None,
			fgcolor = None, bgcolor = None, label_color = None,
			hlfg = None, hlbg = None):
		""" align_around: label and value are aligned either side of this.
			Derived classes must provide method
			render_value(surf, x, y, fg, bg, highlighted)
			which returns height of rendered item.
		"""
		if not label_color:
			label_color = (255, 40, 40)
		self.label_color = label_color
		self.align_around = align_around
		MenuItem.__init__(self, self.handler, label, fgcolor, bgcolor,
			hlfg, hlbg)
		self.label_height = FONT_SIZE

	def render(self, surf, y, highlighted):
		if highlighted:
			fgcolor = self.hlfg
			bgcolor = self.hlbg
			label_fg = self.label_color
		else:
			if self.sensitive:
				fgcolor = self.fgcolor
				label_fg = self.label_color
			else:
				fgcolor = self.hlbg
				label_fg = self.hlbg
			bgcolor = self.bgcolor
		if self.align_around:
			xmid = self.align_around
		else:
			xmid = surf.get_width() / 2
		lsurf = _render_text(self.label, label_fg, self.bgcolor, FONT_SIZE)
		self.label_height = lsurf.get_height()
		surf.blit(lsurf, (xmid - CENTRE_GAP / 2 - lsurf.get_width(), y))
		#vsurf = _render_text(self.label, fgcolor, bgcolor, FONT_SIZE)
		return max(self.render_value(surf, xmid + CENTRE_GAP / 2, y,
			fgcolor, bgcolor, highlighted), self.label_height)
	
	def handler(self):
		pass


class OptionMenuItemBase(LabelledMenuItem):
	""" Cycles through a number of values with arrow keys. Base class"""
	def __init__(self, handler, label, index, value_names, value_values,
			align_around = None, fgcolor = None, bgcolor = None,
			label_color = None, hlfg = None, hlbg = None):
		""" handler gets called with value of selected item
			value_names is a list of possible values as shown in menu.
			value_values is correspond list of values as passed to options
			handler.
			align_around: label and value are aligned either side of this
		"""
		if not label_color:
			label_color = (255, 40, 40)
		self.label_color = label_color
		self.index = index
		self.value_names = value_names
		self.value_values = value_values
		self.client_handler = handler
		LabelledMenuItem.__init__(self, label, align_around, fgcolor, bgcolor,
			hlfg, hlbg)

	def render_value(self, surf, x, y, fgcolor, bgcolor, highlighted):
		if self.index >= len(self.value_names):
			index = 0
		else:
			index = self.index
		return render_text_at(surf, self.get_value_to_display(index),
				x, y, fgcolor, bgcolor, FONT_SIZE)[1]

	def get_value_to_display(self, index):
		""" This may be overridden to temporarily display a different value. """
		return self.value_names[index]
	
	# Derived classes may set defer_update in case they need to call
	# updated() after their handler. The windowed/full-screen option
	# needs to do this to make sure the menu is redrawn after changing
	# the display mode
	def inc(self, defer_update = False):
		if self.index < len(self.value_values) - 1:
			self.index += 1
		if not defer_update:
			self.updated()

	def dec(self, defer_update = False):
		if self.index > 0:
			self.index -= 1
		if not defer_update:
			self.updated()
	
	def call_handler(self):
		self.client_handler(self.value_values[self.index])


class OptionMenuItem(OptionMenuItemBase):
	""" Cycles through a number of values with left/right keys. Return is
	like right with wrap-around. handler is called at every change. """
	def activate(self):
		if self.index < len(self.value_values) - 1:
			self.index += 1
		else:
			self.index = 0
		self.updated()
		self.call_handler()
	
	def inc(self):
		OptionMenuItemBase.inc(self, True)
		self.call_handler()
		self.updated()

	def dec(self):
		OptionMenuItemBase.dec(self, True)
		self.call_handler()
		self.updated()


class ActivatableOptionMenuItem(OptionMenuItemBase):
	""" Cycles through a number of values with left/right keys.
	Return activates. handler is only called at activation. """
	def activate(self):
		self.call_handler()


class BoolMenuItem(OptionMenuItem):
	def __init__(self, handler, label, value, value_names,
			fgcolor = None, bgcolor = None,
			label_color = None, hlfg = None, hlbg = None):
		OptionMenuItem.__init__(self, handler, label, value,
				value_names, (0, 1),
				fgcolor, bgcolor, label_color, hlfg, hlbg)

SLIDER_BORDER = 2
SLIDER_TROUGH = 2

class SliderMenuItem(LabelledMenuItem):
	""" A range of values represented as a slider. """
	def __init__(self, label, min, max, step, start_value, updated_handler,
			align_around = None,
			fgcolor = None, bgcolor = None, label_color = None):
		self.min = min
		self.max = max
		self.step = step
		self.value = start_value
		self.updated_handler = updated_handler
		LabelledMenuItem.__init__(self, label, align_around,
				fgcolor, bgcolor, label_color)

	def render_value(self, surf, x, y, fgcolor, bgcolor, highlighted):
		self.value = float(self.value)
		if highlighted:
			border_color = self.fgcolor
		elif self.sensitive:
			border_color = self.label_color
		else:
			border_color = self.hlbg
		xmax = surf.get_width() * 0.75
		y += int((self.label_height - FONT_SIZE) / 2)
		pygame.draw.rect(surf, border_color, (x, y, xmax - x, FONT_SIZE),
				SLIDER_BORDER)
		if self.value != self.min:
			proportion = (self.value - self.min) / (self.max - self.min)
			pygame.draw.rect(surf, self.hlbg,
					(x + SLIDER_BORDER + SLIDER_TROUGH,
						y + SLIDER_BORDER + SLIDER_TROUGH,
						(xmax - x - 2 * (SLIDER_BORDER + SLIDER_TROUGH)) \
								* proportion + 1,
						FONT_SIZE - 2 * (SLIDER_BORDER + SLIDER_TROUGH) + 1),
					0)
		return self.label_height

	def inc(self):
		val = self.value + self.step
		if val < self.max:
			self.value = val
		else:
			self.value = self.max
		self.updated()
		self.updated_handler(self.value)

	def dec(self):
		val = self.value - self.step
		if val > self.min:
			self.value = val
		else:
			self.value = self.min
		self.updated()
		self.updated_handler(self.value)

