import pygame

from bombz import bitmaps, board, controls, sounds

# Must be factor of board.TILE_SIZE!
PIXELS_PER_TICK = 4

PLAYING = 0
COMPLETED = 1
COMPLETED_ALL = 2
COMPLETED_OUT_OF_TIME_EASY = 3
COMPLETED_OUT_OF_TIME_MEDIUM = 4
BLOWN_UP = 5
OUT_OF_TIME = 6
ABORTED = 7
N_OUTCOMES = 8

class PusherBitmaps:
	def __init__(self):
		self.up = bitmaps.get_bitmap("manup")
		self.down = bitmaps.get_bitmap("mandown")
		self.left = bitmaps.get_bitmap("manleft")
		self.right = bitmaps.get_bitmap("manright")
		self.blank = bitmaps.get_bitmap("blank")

class Pusher:
	def Reset(self):
		self.x = 0
		self.y = 0
		self.dx = 0
		self.dy = 0
		self.offset = 0
		self.was_vert = False
		self.bitmap = self.bitmaps.up
		self.has_match = False
		self.pushing = board.CELL_EMPTY
		self.completed = False
		self.blownup = False
		self.old_pushing = board.CELL_EMPTY
		self.old_dx = 0
		self.old_dy = 0

	def __init__(self):
		self.bitmaps = PusherBitmaps()
		self.Reset()

	def NewBoard(self, brd, stats):
		self.Reset()
		self.board = brd
		self.stats = stats
		(self.x, self.y) = brd.GetStartPos()

	def CanDetonate(self, content):
		return self.has_match and board.IsUnfusedBomb(content)

	def CheckMoveValid(self, dx, dy):
		moveto = self.board.GetCellAt(self.x, \
				self.y).GetNeighbourContent(dx, dy)
		if moveto == None:
			# Trying to move off screen
			if self.board.GotAllBombs():
				return True
			else:
				return False
		if moveto <= board.CELL_EARTH2 or moveto == board.CELL_MATCH:
			return True
		if not self.has_match and board.IsPushable(moveto):
			if self.board.GetCellAt(self.x, \
					self.y).GetNeighbourContent(2 * dx, 2 * dy) \
					== board.CELL_EMPTY:
				return True
		elif self.CanDetonate(moveto):
			return True
		return False

	def GetNewDirection(self):
		[left, right, up, down] = controls.get_direction()
		if left and not self.CheckMoveValid(-1, 0):
			left = False
		if right and not self.CheckMoveValid(1, 0):
			right = False
		if up and not self.CheckMoveValid(0, -1):
			up = False
		if down and not self.CheckMoveValid(0, 1):
			down = False
		# If trying to go diagonally use was_vert flag to alternate between
		# vert and horiz
		if (left or right) and (up or down):
			if self.was_vert:
				up = False
				down = False
			else:
				left = False
				right = False
		if left:
			self.dx = -1
			self.dy = 0
			self.was_vert = False
			self.bitmap = self.bitmaps.left
		elif right:
			self.dx = 1
			self.dy = 0
			self.was_vert = False
			self.bitmap = self.bitmaps.right
		elif up:
			self.dx = 0
			self.dy = -1
			self.was_vert = True
			self.bitmap = self.bitmaps.up
		elif down:
			self.dx = 0
			self.dy = 1
			self.was_vert = True
			self.bitmap = self.bitmaps.down
		else:
			self.dx = 0
			self.dy = 0
		new_cell = self.board.GetCellAt(self.x + self.dx, self.y + self.dy)
		if not new_cell:
			self.dx = 0
			self.dy = 0
			self.completed = True
			return
		self.pushing = new_cell.GetContent()
		if self.CanDetonate(self.pushing):
			sounds.fuse.play()
			if self.pushing == board.CELL_BOMB1:
				new_cell.SetContent(self.board.bomb1_fused_first)
			else:
				new_cell.SetContent(self.board.bomb2_fused_first)
			self.pushing = board.CELL_EMPTY
			self.dx = 0
			self.dy = 0
			self.has_match = False
		elif board.IsPushable(self.pushing):
			new_cell.SetContent(board.CELL_EMPTY)
		else:
			self.pushing = board.CELL_EMPTY
	
	def RenderBitmap(self, surface, bitmap, dx = 0, dy = 0):
		surface.blit(bitmap, \
				((self.x + dx) * board.TILE_SIZE + self.offset * self.dx, \
				(self.y + dy) * board.TILE_SIZE + self.offset * self.dy))
	
	def RenderBlank(self, surface):
		return
		self.RenderBitmap(surface, self.bitmaps.blank)
		if self.pushing:
			self.RenderBitmap(surface, self.bitmaps.blank, self.dx, self.dy)
	
	def RenderBody(self, surface):
		self.RenderBitmap(surface, self.bitmap)
		if self.pushing:
			self.RenderBitmap(surface, \
					board.GetBitmapForContent(self.pushing), self.dx, self.dy)
	
	def Update(self):
		if not self.offset:
			self.GetNewDirection()
		if self.dx or self.dy:
			self.offset += PIXELS_PER_TICK
			if self.offset == board.TILE_SIZE:
				self.offset = 0
				self.x += self.dx
				self.y += self.dy
				if self.board.GetCellAt(self.x, self.y).GetContent() \
						== board.CELL_MATCH:
					sounds.match.play()
					self.has_match = True
				if self.pushing:
					self.board.SetContentAt(self.x + self.dx, \
							self.y + self.dy, self.pushing)
				self.board.SetContentAt(self.x, self.y, board.CELL_EMPTY)
		x0 = self.x - 1
		x1 = self.x + 1
		y0 = self.y - 1
		y1 = self.y + 1

		if self.dx != self.old_dx or self.dy != self.old_dy or \
				self.pushing != self.old_pushing:
			if self.pushing != board.CELL_EMPTY:
				# Pushing indicates more difficulty so count it as 2 moves
				self.stats.move_made(2)
			else:
				self.stats.move_made(1)
		self.old_dx = self.dx
		self.old_dy = self.dy
		self.old_pushing = self.pushing

		if self.pushing:
			if self.dx == 1:
				x1 += 1
			elif self.dx == -1:
				x0 -= 1
			elif self.dy == 1:
				y1 += 1
			elif self.dy == -1:
				y0 -= 1
		[x0, y0, x1, y1] = board.ClipRect(x0, y0, x1, y1)
		for y in range(y0, y1 + 1):
			for x in range(x0, x1 + 1):
				if board.IsExplosion(self.board.GetContentAt(x, y)):
					self.blownup = True
		if self.blownup:
			return BLOWN_UP
		elif self.completed:
			return COMPLETED
		else:
			return PLAYING
