"""Find directories for storing config and data in in a portable way

Functions:
	user_data_dir(...)	path to user's app-specific data dir 
	user_conf_dir(...)	path to user's app-specific config dir

	May both be same directory on some platforms

	url		= URL belonging to organisation owning app
	appname	= Application name
	create	= Whether to create directory
"""

import sys
import os

from bombz import appdir

def _get_xdg_dir(confordata, url, appname):
	basedir = os.environ.get("XDG_" + confordata + "_HOME")
	if not basedir:
		if confordata == "CONFIG":
			basedir = "~/.config"
		elif confordata == "CACHE":
			basedir = "~/.cache"
		else:
			basedir = "~/.local/share"
	basedir = os.path.expanduser(basedir)
	return os.path.join(basedir, url, appname)

def _get_mac_dir(confordata, url, appname):
	elements = url.split(".")
	elements.reverse()
	return os.path.expanduser(os.path.join("~/Library", confordata,
		'.'.join(elements), appname))

def _get_win_dir(confordata, url, appname):
	# confordata is only Win 9x; they'll both be in the same place for NT
	basedir = os.environ.get("USERPROFILE")
	if not basedir:
		# Must be Win 9x, resort to installation dir
		return os.path.join(appdir.appdir, confordata)
	return os.path.join(basedir, "Application Data", url, appname)

def ensure_dir(dirname):
	if not os.path.exists(dirname):
		os.makedirs(dirname, 0755)

def _user_whatever_dir(url, appname, win_sub_dir, mac_sub_dir, xdg_var_stub,
		create = True):
	if sys.platform.startswith("win"):
		path = _get_win_dir(win_sub_dir, url, appname)
	elif sys.platform == 'darwin':
		path = _get_mac_dir(mac_sub_dir, url, appname)
	# FIXME: Include RISC OS?
	else:
		# Assume Linux/Unix
		path = _get_xdg_dir(xdg_var_stub, url, appname)
	if create:
		ensure_dir(path)
	return path

def user_data_dir(url, appname, create = True):
	return _user_whatever_dir(url, appname, "Data", "Application Support",
			"DATA", create)

def user_conf_dir(url, appname, create = True):
	return _user_whatever_dir(url, appname, "Config", "Preferences",
			"CONFIG", create)

def user_cache_dir(url, appname, create = True):
	return _user_whatever_dir(url, appname, "Cache", "Caches",
			"CACHE", create)


class UserDirs:
	""" Wrapper class so you only need to pass url and appname once. """
	def __init__(self, url, appname):
		self.conf_dir = user_conf_dir(url, appname, True)
		self.data_dir = user_data_dir(url, appname, True)
		self.cache_dir = user_cache_dir(url, appname, True)
	def data(self):
		return self.data_dir
	def conf(self):
		return self.conf_dir
	def cache(self):
		return self.cache_dir

dirs = None
